class Job:
    """Generic job class."""

    def __init__(self, namespace=None, name=None, image=None, data=None, **kw):
        if data is not None:
            self._raw = data
        else:
            self._raw = self.create(
                namespace,
                name=name,
                image=image,
                **kw)

    @property
    def active(self):
        raise NotImplementedError()

    @property
    def finished(self):
        raise NotImplementedError()

    @property
    def id(self):
        raise NotImplementedError()

    def get_payload(self):
        raise NotImplementedError()

    def payload(self):
        return self._raw



class NodeList:
    """Generic Node class."""

    def __init__(self, data=None, **kw):
        self._raw = data
        self._nodes = []

    def __len__(self):
        return len(self.values())

    def __iter__(self):
        for node in self.values():
            yield node

    def __setitem__(self, key, value):
        self._nodes[key] = value

    def __getitem__(self, key):
        return self._nodes[key]

    @property
    def total(self):
        return len(self._nodes)

    def values(self):
        return self._nodes



from aioclustermanager.k8s.caller import K8SCaller
from base64 import b64decode

import aiohttp
import asyncio
import logging
import os
import ssl
import tempfile

logger = logging.getLogger('aioclustermanager')

SERVICE_HOST_ENV_NAME = "KUBERNETES_SERVICE_HOST"
SERVICE_PORT_ENV_NAME = "KUBERNETES_SERVICE_PORT"
SERVICE_TOKEN_ENV_NAME = "KUBERNETES_SERVICE_TOKEN"
SERVICE_TOKEN_FILENAME = "/var/run/secrets/kubernetes.io/serviceaccount/token"
SERVICE_CERT_FILENAME = "/var/run/secrets/kubernetes.io/serviceaccount/ca.crt"


def _join_host_port(host, port):
    template = "%s:%s"
    host_requires_bracketing = ':' in host or '%' in host
    if host_requires_bracketing:
        template = "[%s]:%s"
    return template % (host, port)


class Configuration:
    file = None
    ssl_context = None
    cert_file = None

    def __init__(self, environment, loop=None):
        self.headers = {}
        if loop is None:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = loop

        self.environment = environment

        if environment.get('in_cluster'):
            if SERVICE_TOKEN_ENV_NAME in os.environ:
                token = os.environ[SERVICE_TOKEN_ENV_NAME]
            else:
                with open(SERVICE_TOKEN_FILENAME) as fi:
                    token = fi.read()
            self.headers = {
                'Authorization': 'Bearer ' + token
            }
            environment.update({
                'skip_ssl': True,
                'endpoint': _join_host_port(
                    os.environ[SERVICE_HOST_ENV_NAME],
                    os.environ[SERVICE_PORT_ENV_NAME])
            })

        if environment.get('certificate') is not None:
            # Certificate management
            self.load_certificate()
        elif environment.get('certificate_file') is not None:
            self.load_certificate_file()
        elif environment.get('user') and environment.get('credentials'):
            self.load_basic_auth()
        elif environment.get('skip_ssl'):
            self.load_skip_ssl()

        ssl_context = None
        if environment.get('ca') is not None:
            self.file = tempfile.NamedTemporaryFile(delete=False)
            self.file.write(bytes(environment['ca'], encoding='utf-8'))
            self.file.close()
            self.ssl_context = ssl.SSLContext()
            ssl_context.load_verify_locations(self.file.name)
        elif environment.get('ca_file') is not None:
            self.ssl_context = ssl.SSLContext()
            self.ssl_context.load_verify_locations(environment['ca_file'])

        if environment['skip_ssl']:
            self.verify = False
        else:
            self.verify = True

    def load_skip_ssl(self):
        self.session = aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(verify_ssl=False, loop=self.loop),
            headers=self.headers, loop=self.loop)

    def load_basic_auth(self):
        basic_auth = aiohttp.BasicAuth(
            self.environment['user'], self.environment['credentials'])
        self.session = aiohttp.ClientSession(
            auth=basic_auth, headers=self.headers, loop=self.loop)

    def load_certificate_file(self):
        logger.debug('Loading cert files')
        ssl_client_context = ssl.create_default_context(
            purpose=ssl.Purpose.CLIENT_AUTH)
        if 'key_file' in self.environment:
            ssl_client_context.load_cert_chain(
                certfile=self.environment['certificate_file'],
                keyfile=self.environment['key_file'])
        else:
            ssl_client_context.load_cert_chain(
                certfile=self.environment['certificate_file'])
        conn = aiohttp.TCPConnector(
            ssl_context=ssl_client_context, loop=self.loop)
        self.session = aiohttp.ClientSession(
            connector=conn, headers=self.headers, loop=self.loop)

    def load_certificate(self):
        ssl_client_context = ssl.create_default_context(
            purpose=ssl.Purpose.CLIENT_AUTH)
        self.cert_file = tempfile.NamedTemporaryFile(delete=False)
        self.cert_file.write(b64decode(self.environment['certificate']))
        self.cert_file.close()
        self.client_key = tempfile.NamedTemporaryFile(delete=False)
        self.client_key.write(b64decode(self.environment['key']))
        self.client_key.close()

        ssl_client_context.load_cert_chain(
            certfile=self.cert_file.name, keyfile=self.client_key.name)
        conn = aiohttp.TCPConnector(
            ssl_context=ssl_client_context, loop=self.loop)
        self.session = aiohttp.ClientSession(
            connector=conn, loop=self.loop, headers=self.headers)


class K8SContextManager:
    def __init__(self, environment, loop=None):
        self.environment = environment
        if loop is None:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = loop

    async def __aenter__(self):
        return await self.open()

    async def open(self):
        self.config = Configuration(self.environment, self.loop)
        return K8SCaller(
            self.config.ssl_context,
            self.environment['endpoint'],
            self.config.session,
            verify=self.config.verify)

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    async def close(self):
        if self.config.file is not None:
            os.unlink(self.config.file.name)
        await self.config.session.close()


async def create_k8s_caller(environment):
    config = Configuration(environment)
    return K8SCaller(
        config.ssl_context,
        environment['endpoint'],
        config.session,
        verify=config.verify)



from aioclustermanager.job import Job
from copy import deepcopy

import json

K8S_JOB = {
    "kind": "Job",
    "metadata": {
        "name": "",
        "namespace": ""
    },
    "spec": {
        "backoffLimit": 1,
        "restartPolicy": "Never",
        "template": {
            "metadata": {
                "name": ""
            },
            "spec": {
                "containers": [
                    {
                        "name": "",
                        "image": "",
                        "resources": {
                            "limits": {
                            }
                        },
                        "imagePullPolicy": "IfNotPresent"
                    }
                ],
                "restartPolicy": "Never"
            }
        }
    }
}


class K8SJob(Job):
    @property
    def active(self):
        status = self._raw['status']
        return False if 'active' not in status else status['active']

    @property
    def finished(self):
        status = self._raw['status']
        return 'failed' in status or 'succeeded' in status

    @property
    def id(self):
        return self._raw['metadata']['name']

    @property
    def command(self):
        return self._raw['spec']['template']['spec']['containers'][0]['command']  # noqa

    @property
    def image(self):
        return self._raw['spec']['template']['spec']['containers'][0]['image']

    def create(self, namespace, name, image, **kw):
        job_info = deepcopy(K8S_JOB)
        job_info['metadata']['name'] = name
        job_info['metadata']['namespace'] = namespace
        job_info['spec']['template']['metadata']['name'] = name
        job_info['spec']['template']['spec']['containers'][0]['name'] = name
        job_info['spec']['template']['spec']['containers'][0]['image'] = image

        if 'entrypoint' in kw and kw['entrypoint'] is not None:
            job_info['spec']['template']['spec']['containers'][0]['entrypoint'] = kw['entrypoint']  # noqa

        if 'command' in kw and kw['command'] is not None:
            job_info['spec']['template']['spec']['containers'][0]['command'] = kw['command']  # noqa

        if 'args' in kw and kw['args'] is not None:
            job_info['spec']['template']['spec']['containers'][0]['args'] = kw['args']  # noqa

        if 'mem_limit' in kw and kw['mem_limit'] is not None:
            job_info['spec']['template']['spec']['containers'][0]['resources']['limits']['memory'] = kw['mem_limit']  # noqa

        if 'cpu_limit' in kw and kw['cpu_limit'] is not None:
            job_info['spec']['template']['spec']['containers'][0]['resources']['limits']['cpu'] = kw['cpu_limit']  # noqa

        if 'volumes' in kw and kw['volumes'] is not None:
            job_info['spec']['template']['spec']['volumes'] = kw['volumes']

        if 'volumeMounts' in kw and kw['volumeMounts'] is not None:
            job_info['spec']['template']['spec']['containers'][0]['volumeMounts'] = kw['volumeMounts']  # noqa

        if 'envFrom' in kw and kw['envFrom'] is not None:
            job_info['spec']['template']['spec']['containers'][0]['envFrom'] = kw['envFrom']  # noqa

        if 'envvars' in kw and kw['envvars'] is not None:
            envlist = []
            for key, value in kw['envvars'].items():
                envlist.append({
                    "name": key,
                    "value": value
                })
            job_info['spec']['template']['spec']['containers'][0]['env'] = envlist  # noqa

        return job_info

    def get_payload(self):
        container = self._raw['spec']['template']['spec']['containers'][0]
        for env in container['env']:
            if env['name'] == 'PAYLOAD':
                data = env['value']
                return json.loads(data)
        return None



from pathlib import Path

import os
import yaml


def get_k8s_config():
    home = str(Path.home())
    # Designed to run on docker desktop k8s support
    # XXX DOES NOT CURRENTLY WORK WITH containers running in k8s already
    with open(home + '/.kube/config', 'r') as f:
        configuration = yaml.load(f)

    CERT_DOCKER = None
    KEY_DOCKER = None
    CERT_DOCKER_FILE = None
    KEY_DOCKER_FILE = None
    K8S_CA = None
    K8S_CA_FILE = None
    K8S_USER = None
    K8S_PASSWORD = None
    K8S_SKIP_SSL = True
    K8S_ENDPOINT = 'localhost:6443'
    TRAVIS = os.environ.get('TRAVIS', 'false')
    # Looking for docker-for-desktop or minikube
    defined = False
    for user in configuration['users']:
        if user['name'] == 'docker-for-desktop':
            CERT_DOCKER = user['user']['client-certificate-data']
            KEY_DOCKER = user['user']['client-key-data']
            defined = True
        if user['name'] == 'minikube' and defined is False and TRAVIS != 'true':  # noqa
            CERT_DOCKER_FILE = user['user']['client-certificate']
            KEY_DOCKER_FILE = user['user']['client-key']
        if TRAVIS == 'true':
            K8S_USER = 'testinguser'
            K8S_PASSWORD = '12345678'
    for cluster in configuration['clusters']:
        if cluster['name'] == 'minikube' and defined is False:
            K8S_SKIP_SSL = False
            K8S_ENDPOINT = cluster['cluster']['server']
            K8S_ENDPOINT = K8S_ENDPOINT.replace("https://", "")
            K8S_CA_FILE = cluster['cluster']['certificate-authority']

    config_k8s = {
        'user': os.environ.get('TEST_K8S_USER', K8S_USER),
        'credentials': os.environ.get('TEST_K8S_CREDS', K8S_PASSWORD),
        'ca': os.environ.get('TEST_K8S_CA', K8S_CA),
        'ca_file': os.environ.get('TEST_K8S_CA_FILE', K8S_CA_FILE),
        'endpoint': os.environ.get('TEST_K8S_ENDPOINT', K8S_ENDPOINT),
        'skip_ssl': K8S_SKIP_SSL,
        'certificate': os.environ.get('TEST_K8S_CERT', CERT_DOCKER),
        'key': os.environ.get('TEST_K8S_KEY', KEY_DOCKER),
        'certificate_file': os.environ.get('TEST_K8S_CERT_FILE', CERT_DOCKER_FILE),  # noqa
        'key_file': os.environ.get('TEST_K8S_KEY_FILE', KEY_DOCKER_FILE)
    }
    return config_k8s



import pytest
import asyncio
pytestmark = pytest.mark.asyncio


async def test_get_jobs_k8s(kubernetes):
    # We clean up all the jobs on the namespace

    result = await kubernetes.get_nodes()
    assert len(result) > 0

    result = await kubernetes.cleanup_jobs('aiocluster-test')
    assert result == 0

    jobs_info = await kubernetes.list_jobs('aiocluster-test')
    assert jobs_info.total == 0

    await kubernetes.create_job(
        'aiocluster-test',  # namespace
        'test-job',  # jobid
        'perl',  # image
        command=["perl", "-Mbignum=bpi", "-wle", "print bpi(2000)"],
    )

    job_info = await kubernetes.get_job('aiocluster-test', 'test-job')
    assert job_info.id == 'test-job'

    jobs_info = await kubernetes.list_jobs('aiocluster-test')
    assert jobs_info.total == 1

    result = await kubernetes.wait_for_job('aiocluster-test', 'test-job')
    assert result == 1

    job_info = await kubernetes.get_job('aiocluster-test', 'test-job')
    assert job_info.finished
    assert job_info.id == 'test-job'

    executions = await kubernetes.list_job_executions(
        'aiocluster-test', 'test-job')
    assert len(executions) > 0

    log = await kubernetes.get_execution_log(
        'aiocluster-test', 'test-job', executions[0].internal_id)
    assert "3.14" in log

    result = await kubernetes.delete_job('aiocluster-test', 'test-job')
    assert result is True

    job_info = await kubernetes.get_job('aiocluster-test', 'test-job')
    assert job_info is None


async def test_get_jobs_limit_k8s(kubernetes):
    # We clean up all the jobs on the namespace

    result = await kubernetes.define_quota(
        'aiocluster-test',
        cpu_limit="400m", mem_limit="900M")
    assert result is True

    result = await kubernetes.create_job(
        'aiocluster-test',  # namespace
        'test-job',  # jobid
        'tensorflow/tensorflow:latest-py3',  # image
        cpu_limit="300m",
        mem_limit="800M",
        command=["jupyter-nbconvert"],
        args=[
            "--execute", "3_mnist_from_scratch.ipynb",
            "--ExecutePreprocessor.timeout=380"],
    )
    assert result is True

    await asyncio.sleep(5)

    result = await kubernetes.create_job(
        'aiocluster-test',  # namespace
        'test-job-2',  # jobid
        'tensorflow/tensorflow:latest-py3',  # image
        cpu_limit="300m",
        mem_limit="800M",
        command=["jupyter-nbconvert"],
        args=[
            "--execute", "3_mnist_from_scratch.ipynb",
            "--ExecutePreprocessor.timeout=380"],
    )
    assert result is True

    await asyncio.sleep(20)

    # we want to wait that the first job starts

    result = await kubernetes.wait_for_job_execution_status(
        'aiocluster-test', 'test-job')
    assert result == 'Running'

    result = await kubernetes.waiting_jobs('aiocluster-test')
    assert len(result) == 1
    assert result[0] == 'test-job-2'

    result = await kubernetes.running_jobs('aiocluster-test')
    assert len(result) == 1
    assert result[0] == 'test-job'

    result = await kubernetes.delete_job('aiocluster-test', 'test-job')
    assert result is True

    await asyncio.sleep(50)

    result = await kubernetes.waiting_jobs('aiocluster-test')
    assert len(result) == 0

    result = await kubernetes.running_jobs('aiocluster-test')
    assert len(result) == 1
    assert result[0] == 'test-job-2'


async def test_get_tfjobs_k8s(kubernetes):
    # We clean up all the jobs on the namespace

    result = await kubernetes.define_quota(
        'aiocluster-test',
        cpu_limit="400m", mem_limit="900M")
    assert result is True

    await kubernetes.install_tfjobs('aiocluster-test')

    result = await kubernetes.create_tfjob(
        'aiocluster-test',  # namespace
        'test-tfjob',  # jobid
        'gcr.io/tf-on-k8s-dogfood/tf_sample:dc944ff',  # image
        cpu_limit="300m",
        mem_limit="800M",
        workers=1, ps=2, masters=1
    )
    assert result is True

    await kubernetes.get_tfjob('aiocluster-test', 'test-tfjob')

    await kubernetes.list_tfjobs('aiocluster-test')



from aioclustermanager.executions_list import ExecutionList
from aioclustermanager.nomad.execution import NomadExecution


class NomadExecutionList(ExecutionList):

    def process(self):
        for pod in self._raw:
            self._executions.append(NomadExecution(data=pod))

    def statuses(self):
        return [execution.status for execution in self]

    def has_failed_anytime(self):
        for execution in self:
            if execution.failed:
                return True
        return False

    def is_running(self):
        return self[-1].running

    def is_done(self):
        if len(self) > 0:
            return self[-1].finished
        return True



from aioclustermanager.node import Node


class K8SNode(Node):
    @property
    def id(self):
        return self._raw['metadata']['name']

    @property
    def hostname(self):
        return self._raw['metadata']['labels']['kubernetes.io/hostname']



from aioclustermanager.execution import Execution
from aioclustermanager.k8s.const import FAILED, PENDING, RUNNING


class K8SExecution(Execution):

    @property
    def pending(self):
        if self._raw['status']['phase'] == PENDING:
            return True
        else:
            return False

    @property
    def running(self):
        if self._raw['status']['phase'] == RUNNING:
            return True
        else:
            return False

    @property
    def failed(self):
        if self._raw['status']['phase'] == FAILED:
            return True
        else:
            return False

    @property
    def status(self):
        return self._raw['status']['phase']

    @property
    def events(self):
        result = []
        for condition in self._raw['status']['conditions']:
            result.append((condition['type'], condition['lastTransitionTime']))
        return result

    @property
    def id(self):
        return self._raw['metadata']['name']

    @property
    def internal_id(self):
        return self._raw['metadata']['name']



from aioclustermanager.k8s import K8SContextManager
from aioclustermanager.manager import ClusterManager
from aioclustermanager.nomad import NomadContextManager
from aioclustermanager.tests.utils import get_k8s_config

import os
import pytest


@pytest.fixture(scope='function')
async def k8s_config():
    return get_k8s_config()


@pytest.fixture(scope='function')
async def nomad_config():
    config_nomad = {
        'endpoint': os.environ.get('TEST_NOMAD_ENDPOINT', 'localhost:4646')
    }
    return config_nomad


@pytest.fixture(scope='function')
async def kubernetes(k8s_config):

    async with K8SContextManager(k8s_config) as context:
        cm = ClusterManager(context)
        await cm.delete_namespace('aiocluster-test')
        await cm.create_namespace('aiocluster-test')
        yield cm
        await cm.delete_namespace('aiocluster-test')


@pytest.fixture(scope='function')
async def nomad(nomad_config):

    async with NomadContextManager(nomad_config) as context:
        cm = ClusterManager(context)
        await cm.delete_namespace('aiocluster-test')
        await cm.create_namespace('aiocluster-test')
        yield cm
        await cm.delete_namespace('aiocluster-test')




from aioclustermanager.node_list import NodeList
from aioclustermanager.nomad.node import NomadNode


class NomadNodeList(NodeList):

    def __init__(self, data=None, **kw):
        self._raw = data
        self._nodes = []
        for node in self._raw:
            self._nodes.append(NomadNode(data=node))



from aioclustermanager.execution import Execution
from aioclustermanager.nomad.const import FAILED, PENDING, RUNNING, SUCCEEDED


class NomadExecution(Execution):

    @property
    def pending(self):
        if self.status == PENDING:
            return True
        else:
            return False

    @property
    def running(self):
        if self.status == RUNNING:
            return True
        else:
            return False

    @property
    def failed(self):
        if self.status == FAILED:
            return True
        else:
            return False

    @property
    def finished(self):
        if self.status == SUCCEEDED:
            return True
        else:
            return False

    @property
    def status(self):
        return self._raw['ClientStatus']

    @property
    def events(self):
        result = []
        for condition in self._raw['TaskStates'][self.id]['Events']:
            result.append((condition['Type'], condition['Time']))
        return result

    @property
    def id(self):
        return self._raw['JobID']

    @property
    def internal_id(self):
        return self._raw['ID']



class Node:
    """Generic node class."""

    def __init__(self, data=None, **kw):
        self._raw = data

    @property
    def id(self):
        raise NotImplementedError()

    @property
    def hostname(self):
        raise NotImplementedError()



class JobList:
    """Generic job class."""

    def __init__(self, data=None, **kw):
        self._raw = data

    @property
    def values(self):
        raise NotImplementedError()

    def __len__(self):
        return len(self.values())

    def __iter__(self):
        for job in self.values():
            yield job

    def schedulled_pods(self):
        for key, value in self.items():
            if len(value) > 0:
                # XXX work with both nomad and k8s?
                schedulled = list(filter(lambda x: x[0] == 'PodScheduled', value[0].events))  # noqa



from aioclustermanager.job_list import JobList
from aioclustermanager.k8s.tf_job import K8STFJob


class K8STFJobList(JobList):

    def __init__(self, data=None, **kw):
        self._raw = data
        self._jobs = []
        for job in self._raw['items']:
            self._jobs.append(K8STFJob(data=job))

    @property
    def total(self):
        return len(self._jobs)

    def values(self):
        return self._jobs



from aioclustermanager.k8s import const
from aioclustermanager.k8s.delete import K8SDelete
from aioclustermanager.k8s.executions_list import K8SExecutionList
from aioclustermanager.k8s.job import K8SJob
from aioclustermanager.k8s.job_list import K8SJobList
from aioclustermanager.k8s.namespace import K8SNamespace
from aioclustermanager.k8s.node_list import K8SNodeList
from aioclustermanager.k8s.quota import K8SQuota
from aioclustermanager.k8s.tf_job import K8STFJob
from aioclustermanager.k8s.tf_job_list import K8STFJobList

import concurrent
import json
import logging

logger = logging.getLogger(__name__)

WATCH_OPS = {
    'namespace': 'https://{endpoint}/api/v1/watch/namespaces/{namespace}',
    'job': 'https://{endpoint}/apis/batch/v1/watch/namespaces/{namespace}/jobs/{name}',  # noqa
    'tfjob':
        'https://{endpoint}/apis/kubeflow.org/v1alpha1/watch/namespaces/{namespace}/tfjobs/{name}'  # noqa
}

GET_OPS = {
    'namespace':
        'https://{endpoint}/api/v1/namespaces/{namespace}',
    'list_jobs':
        'https://{endpoint}/apis/batch/v1/namespaces/{namespace}/jobs',
    'job':
        'https://{endpoint}/apis/batch/v1/namespaces/{namespace}/jobs/{name}/status',  # noqa
    'executions':
        'https://{endpoint}/api/v1/namespaces/{namespace}/pods/?labelSelector=job-name={name}',  # noqa
    'tfjob':
        'https://{endpoint}/apis/kubeflow.org/v1alpha1/namespaces/{namespace}/tfjobs/{name}',  # noqa
    'list_tfjobs':
        'https://{endpoint}/apis/kubeflow.org/v1alpha1/namespaces/{namespace}/tfjobs',  # noqa
    'log':
        'https://{endpoint}/api/v1/namespaces/{namespace}/pods/{name}/log',
    'nodes':
        'https://{endpoint}/api/v1/nodes/',
    'configmaps': 'https://{endpoint}/api/v1/namespaces/{namespace}/configmaps?{selector}'  # noqa
}

POST_OPS = {
    'namespace':
        ('https://{endpoint}/api/v1/namespaces', 'v1'),
    'job':
        ('https://{endpoint}/apis/batch/v1/namespaces/{namespace}/jobs', 'batch/v1'),  # noqa
    'tfjob':
        ('https://{endpoint}/apis/kubeflow.org/v1alpha1/namespaces/{namespace}/tfjobs', 'kubeflow.org/v1alpha1'),  # noqa
    'quota':
        ('https://{endpoint}/api/v1/namespaces/{namespace}/resourcequotas', 'v1')  # noqa
}

DELETE_OPS = {
    'namespace':
        ('https://{endpoint}/api/v1/namespaces/{namespace}', 'v1'),
    'job':
        ('https://{endpoint}/apis/batch/v1/namespaces/{namespace}/jobs/{name}', 'batch/v1'),  # noqa
    'tfjob':
        ('https://{endpoint}/apis/kubeflow.org/v1alpha1/namespaces/{namespace}/tfjobs/{name}', 'v1')  # noqa
}


class K8SCaller(object):
    constants = const
    _type = 'k8s'

    def __init__(self, ssl_context, endpoint, session, verify=True):
        self.ssl_context = ssl_context
        self.endpoint = endpoint
        self.session = session
        self.verify = verify

    @property
    def type(self):
        return self._type

    async def get_namespace(self, namespace):
        url = GET_OPS['namespace']
        url = url.format(
            namespace=namespace,
            endpoint=self.endpoint)
        result = await self.get(url)
        if result is None:
            return None
        else:
            return K8SNamespace(data=result)

    async def create_namespace(self, namespace):
        url, version = POST_OPS['namespace']
        url = url.format(
            namespace=namespace,
            endpoint=self.endpoint)
        obj = K8SNamespace(namespace)
        return await self.post(url, version, obj.payload())

    async def delete_namespace(self, namespace):
        url, version = DELETE_OPS['namespace']
        url = url.format(
            namespace=namespace,
            endpoint=self.endpoint)
        obj = K8SDelete()
        return await self.delete(url, version, obj.payload())

    async def wait_added(self, kind, namespace, name=None, timeout=30):
        url = WATCH_OPS[kind]
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        return await self.watch(url, value='ADDED', timeout=timeout)

    async def wait_deleted(self, kind, namespace, name=None, timeout=30):
        url = WATCH_OPS[kind]
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        return await self.watch(url, value='DELETED', timeout=timeout)

    async def define_quota(self, namespace, cpu_limit=None, mem_limit=None):
        url, version = POST_OPS['quota']
        url = url.format(
            namespace=namespace,
            endpoint=self.endpoint)
        obj = K8SQuota(
            namespace=namespace,
            max_cpu=cpu_limit,
            max_memory=mem_limit)
        return await self.post(url, version, obj.payload())

    async def get_nodes(self):
        url = GET_OPS['nodes']
        url = url.format(
            endpoint=self.endpoint)
        result = await self.get(url)
        if result is None:
            return None
        else:
            return K8SNodeList(data=result)

    async def get_job(self, namespace, name):
        url = GET_OPS['job']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        result = await self.get(url)
        if result is None:
            return None
        else:
            return K8SJob(data=result)

    async def get_tfjob(self, namespace, name):
        url = GET_OPS['tfjob']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        result = await self.get(url)
        if result is None:
            return None
        else:
            return K8STFJob(data=result)

    async def get_job_executions(self, namespace, name):
        url = GET_OPS['executions']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        result = await self.get(url)
        if result is None:
            return None
        else:
            return K8SExecutionList(data=result)

    async def get_execution_log(self, namespace, job_id, execution_id):
        url = GET_OPS['log']
        url = url.format(
            namespace=namespace,
            name=execution_id,
            endpoint=self.endpoint)
        result = await self.get(url, noaccept=True)
        if result is None:
            return None
        else:
            return result

    async def get_execution_log_watch(self, namespace, job_id, execution_id):
        url = GET_OPS['log']
        url = url.format(
            namespace=namespace,
            name=execution_id,
            endpoint=self.endpoint)
        async for logline in await self._watch_log(url, timeout=3660):
            yield logline

    async def delete_job(self, namespace, name, wait=False):
        url, version = DELETE_OPS['job']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        obj = K8SDelete()
        await self.delete(url, version, obj.payload())
        if wait:
            return await self.wait_deleted('job', namespace, name)
        return True

    async def delete_tfjob(self, namespace, name, wait=False):
        url, version = DELETE_OPS['tfjob']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        obj = K8SDelete()
        await self.delete(url, version, obj.payload())
        if wait:
            return await self.wait_deleted('tfjob', namespace, name)
        return True

    async def list_jobs(self, namespace):
        url = GET_OPS['list_jobs']
        url = url.format(
            namespace=namespace,
            endpoint=self.endpoint)
        result = await self.get(url)
        if result is None:
            return None
        else:
            return K8SJobList(data=result)

    async def list_tfjobs(self, namespace):
        url = GET_OPS['list_tfjobs']
        url = url.format(
            namespace=namespace,
            endpoint=self.endpoint)
        result = await self.get(url)
        if result is None:
            return None
        else:
            return K8STFJobList(data=result)

    async def create_job(
            self, namespace, name, image,
            command=None, args=None,
            cpu_limit=None, mem_limit=None,
            envvars={}, volumes=None, volumeMounts=None,
            envFrom=None, entrypoint=None):
        url, version = POST_OPS['job']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        obj = K8SJob(
            namespace=namespace,
            name=name,
            image=image,
            command=command, args=args,
            cpu_limit=cpu_limit, mem_limit=mem_limit,
            envvars=envvars, volumes=volumes, volumeMounts=volumeMounts,
            envFrom=envFrom, entrypoint=entrypoint)
        return await self.post(url, version, obj.payload())

    async def create_tfjob(
            self, namespace, name, image,
            command=None, args=None,
            cpu_limit=None, mem_limit=None,
            envvars={}, workers=1, ps=1,
            masters=1, tb_gs=None, volumes=None, volumeMounts=None,
            envFrom=None, entrypoint=None):
        url, version = POST_OPS['tfjob']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        obj = K8STFJob(
            namespace=namespace,
            name=name,
            image=image,
            command=command, args=args,
            cpu_limit=cpu_limit, mem_limit=mem_limit,
            envvars=envvars, workers=workers,
            ps=ps, masters=masters, tb_gs=tb_gs, volumes=volumes,
            volumeMounts=volumeMounts, envFrom=envFrom, entrypoint=entrypoint)
        return await self.post(url, version, obj.payload())

    async def wait_for_job(self, namespace, name):
        url = WATCH_OPS['job']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        try:
            # We only wait 1 hour
            async for data in self._watch(url, 3660):
                json_data = json.loads(data)
                if 'conditions' not in json_data['object']['status']:
                    continue
                for condition in json_data['object']['status']['conditions']:
                    if condition['type'] == 'Complete':
                        return json_data['object']['status']['succeeded']
        except concurrent.futures._base.TimeoutError:
            return None

    # BASIC OPS

    async def _watch(self, url, timeout=20):
        async with self.session.get(
                url + '?watch=true',
                ssl_context=self.ssl_context,
                verify_ssl=self.verify,
                timeout=timeout) as resp:
            assert resp.status == 200
            while True:
                data = await resp.content.readline()
                yield data

    async def _watch_log(self, url, timeout=20):
        async with self.session.get(
                url + '?follow=true',
                ssl_context=self.ssl_context,
                verify_ssl=self.verify,
                timeout=timeout) as resp:
            assert resp.status == 200
            while True:
                data = await resp.content.readline()
                yield data

    async def watch(self, url, value=None, timeout=20):
        try:
            async for data in self._watch(url, timeout):
                json_data = json.loads(data)
                if json_data['type'] == value:
                    break
        except concurrent.futures._base.TimeoutError:
            pass

    async def get(self, url, noaccept=False):
        headers = {}
        if noaccept is False:
            headers['Accept'] = 'application/json'
        async with self.session.get(
                url,
                headers=headers,
                ssl_context=self.ssl_context,
                verify_ssl=self.verify) as resp:
            if resp.status == 404:
                return None
            if resp.status != 200:
                text = await resp.text()
                logger.error('Error: %d - %s' % (resp.status, text))
            assert resp.status == 200
            if noaccept:
                return await resp.text()
            else:
                return await resp.json()

    async def post(self, url, version, payload):
        payload['apiVersion'] = version
        async with self.session.post(
                url,
                json=payload,
                headers={
                    'Accept': 'application/json'
                },
                ssl_context=self.ssl_context,
                verify_ssl=self.verify) as resp:
            if resp.status in [201, 200]:
                return await resp.json()
            else:
                text = await resp.text()
                raise Exception(
                    'Error calling k8s: %d - %s' % (resp.status, text))

    async def delete(self, url, version, payload):
        payload['apiVersion'] = version
        async with self.session.delete(
                url,
                headers={
                    'Accept': 'application/json'
                },
                json=payload,
                ssl_context=self.ssl_context,
                verify_ssl=self.verify) as resp:
            if resp.status == 200:
                return await resp.json()
            elif resp.status == 404:
                return None
            elif resp.status == 409:
                return None
            else:
                text = await resp.text()
                raise Exception(
                    'Error calling k8s: %d - %s' % (resp.status, text))

    async def get_config_maps(self, namespace, labels=None):
        selector = ''
        if labels:
            selector = 'labelSelector={}'.format(
                ','.join(f'{k}={v}' for k, v in labels.items())
            )
        url = GET_OPS['configmaps']
        url = url.format(
            namespace=namespace,
            endpoint=self.endpoint,
            selector=selector)
        return await self.get(url)



import asyncio
import logging

logger = logging.getLogger('aioclustermanager')


class ClusterManager:

    def __init__(self, caller):
        self._caller = caller
        self._const = caller.constants

    @property
    def caller(self):
        return self._caller

    async def create_namespace(self, namespace):
        exist = await self.caller.get_namespace(namespace)
        if exist is None:
            await self.caller.create_namespace(namespace)
            await self.caller.wait_added('namespace', namespace)
            return True
        return False

    async def delete_namespace(self, namespace):
        exist = await self.caller.get_namespace(namespace)
        if exist is not None:
            await self.caller.delete_namespace(namespace)
            await self.caller.wait_deleted('namespace', namespace)
            return True
        return False

    async def get_nodes(self):
        return await self.caller.get_nodes()

    async def define_quota(self, name, cpu_limit=None, mem_limit=None):
        await self.caller.define_quota(name, cpu_limit, mem_limit)
        return True

    async def create_job(
            self, namespace, name, image,
            command=None, args=None,
            cpu_limit=None, mem_limit=None,
            envvars={}, volumes=None, volumeMounts=None,
            envFrom=None, entrypoint=None,
            delete=False, timeout=30, wait=True):
        exist = await self.caller.get_job(namespace, name)
        if exist is not None and delete:
            await self.delete_job(namespace, name, wait=True)
            exist = None

        if exist is None:
            await self.caller.create_job(
                namespace, name, image,
                command=command, args=args,
                cpu_limit=cpu_limit, mem_limit=mem_limit,
                envvars=envvars, volumes=volumes, volumeMounts=volumeMounts,
                envFrom=envFrom, entrypoint=entrypoint)
            if wait:
                await self.caller.wait_added('job', namespace, name=name,
                                             timeout=timeout)
            return True
        return False

    async def list_jobs(self, namespace):
        return await self.caller.list_jobs(namespace)

    async def get_job(self, namespace, name):
        return await self.caller.get_job(namespace, name)

    async def delete_job(self, namespace, name, wait=False):
        return await self.caller.delete_job(namespace, name, wait)

    async def cleanup_jobs(self, namespace):
        jobs = await self.list_jobs(namespace)
        for job in jobs.values():
            await self.delete_job(namespace, job.id, wait=True)
        jobs = await self.list_jobs(namespace)
        return jobs.total

    async def wait_for_job(self, namespace, name):
        return await self.caller.wait_for_job(namespace, name)

    async def list_jobs_executions(self, namespace):
        jobs = await self.list_jobs(namespace)
        result = {}
        for job in jobs:
            result[job.id] = await self.caller.get_job_executions(
                namespace, job.id)
        return result

    async def list_job_executions(self, namespace, job_id):
        return await self.caller.get_job_executions(namespace, job_id)

    async def get_execution_log(self, namespace, job_id, execution_id):
        return await self.caller.get_execution_log(
            namespace, job_id, execution_id)

    async def get_execution_log_watch(self, namespace, job_id, execution_id):
        async for log_line in self.caller.get_execution_log_watch(
                namespace, job_id, execution_id):
            yield log_line

    async def wait_for_job_execution_status(self, namespace, name):
        status = self._const.PENDING
        while status not in [self._const.RUNNING, self._const.ERROR,
                             self._const.SUCCEEDED]:
            await asyncio.sleep(10)
            executions = await self.caller.get_job_executions(namespace, name)
            # It can be multiple executions of a job
            logger.debug(executions.statuses())
            if executions.has_failed_anytime():
                status = self._const.ERROR
            elif executions.is_running():
                status = self._const.RUNNING
            elif executions.is_done():
                status = self._const.SUCCEEDED
        return status

    async def waiting_jobs(self, namespace):
        jobs = await self.list_jobs_executions(namespace)
        result = []
        for key, value in jobs.items():
            if len(value) == 0:
                result.append(key)
        return result

    async def running_jobs(self, namespace):
        jobs = await self.list_jobs_executions(namespace)
        result = []
        for key, value in jobs.items():
            if value.is_running():
                result.append(key)
        return result

    # Only K8S

    async def install_tfjobs(self, namespace, cloud=None, install_helm=False):
        chart = 'https://storage.googleapis.com/tf-on-k8s-dogfood-releases/latest/tf-job-operator-chart-latest.tgz'  # noqa
        call = ['helm', 'install', chart, '-n', 'tf-job', '--wait', '--replace']  # noqa
        if cloud is not None:
            call.extend(['--set', 'cloud=' + cloud])

        check_version = asyncio.create_subprocess_exec(
            'helm', 'version',
            stdout=asyncio.subprocess.PIPE,
        )
        result = await check_version
        stdout = await result.stdout.read()

        if b'Server' not in stdout and install_helm:

            create = asyncio.create_subprocess_exec(
                'helm', 'init',
                stdout=asyncio.subprocess.PIPE,
            )
            await create

        create = asyncio.create_subprocess_exec(
            *call,
            stdout=asyncio.subprocess.PIPE,
        )
        await asyncio.sleep(20)
        await create

    async def create_tfjob(
            self, namespace, name, image,
            command=None, args=None,
            cpu_limit=None, mem_limit=None,
            envvars={}, workers=1, ps=1, masters=1, tb_gs=None,
            delete=False):
        exist = await self.caller.get_tfjob(namespace, name)
        if exist is not None and delete:
            await self.delete_tfjob(namespace, name, wait=True)

        if exist is None:
            await self.caller.create_tfjob(
                namespace, name, image,
                command=command, args=args,
                cpu_limit=cpu_limit, mem_limit=mem_limit,
                envvars=envvars, workers=workers, ps=ps, masters=masters,
                tb_gs=tb_gs)
            await self.caller.wait_added('tfjob', namespace, name=name)
            return True
        return False

    async def get_tfjob(self, namespace, name):
        return await self.caller.get_tfjob(namespace, name)

    async def delete_tfjob(self, namespace, name, wait=False):
        return await self.caller.delete_tfjob(namespace, name, wait)

    async def list_tfjobs(self, namespace):
        return await self.caller.list_tfjobs(namespace)

    async def get_config_maps(self, namespace, labels=None):
        return await self.caller.get_config_maps(namespace, labels or {})



from aioclustermanager.node import Node


class NomadNode(Node):
    @property
    def id(self):
        return self._raw['Name']

    @property
    def hostname(self):
        raise NotImplementedError()



from aioclustermanager.nomad.caller import NomadCaller

import aiohttp
import asyncio


class NomadContextManager:
    def __init__(self, environment, loop=None):
        self.environment = environment
        if loop is None:
            self.loop = asyncio.get_event_loop()
        self.session = None
        self._datacenter = None

    async def __aenter__(self):
        self.session = aiohttp.ClientSession(loop=self.loop)

        url = 'http://{}/v1/agent/self'.format(self.environment['endpoint'])
        async with self.session.get(url) as resp:
            data = await resp.json()
            self._datacenter = data['config']['Datacenter']

        return NomadCaller(
            self.environment['endpoint'],
            self._datacenter,
            self.session)

    async def __aexit__(self, exc_type, exc, tb):
        await self.session.close()


async def create_nomad_caller(environment):
    session = aiohttp.ClientSession()

    url = 'http://{}/v1/agent/self'.format(environment['endpoint'])
    async with session.get(url) as resp:
        data = await resp.json()
        datacenter = data['config']['Datacenter']

    return NomadCaller(
        environment['endpoint'],
        datacenter,
        session)




from aioclustermanager.node_list import NodeList
from aioclustermanager.k8s.node import K8SNode


class K8SNodeList(NodeList):

    def __init__(self, data=None, **kw):
        self._raw = data
        self._nodes = []
        for node in self._raw['items']:
            self._nodes.append(K8SNode(data=node))

    @property
    def total(self):
        return len(self._nodes)

    def values(self):
        return self._nodes
